#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "huffman.h"

huffman_t Remap[100000];

unsigned readBits = 0;
unsigned short readChar = 0;
FILE     *readFile = NULL;

unsigned writeBits = 0;
unsigned writeChar = 0;
FILE *   writeFile = NULL;

int BitsMask[16] = { 0x8000, 0x4000, 0x2000, 0x1000, 0x0800, 0x0400, 0x0200, 0x0100, 0x0080, 0x0040, 0x0020, 0x0010, 0x0008, 0x0004, 0x0002, 0x0001 };

void PrintByte(unsigned val, int size)
{
	int i;

	for( i = size - 1 ; i >= 0 ; i-- )
	{
		if( val & (1 << i) )
		{
			fprintf(stderr, "1");
		}
		else
		{
			fprintf(stderr, "0");
		}
	}
	fprintf(stderr, " ");
}


void SwapShort(unsigned short *s)
{
	union
	{
		unsigned short s;
		unsigned char  c[2];
	}avant, apres;

	avant.s = *s;
	apres.c[0] = avant.c[1];
	apres.c[1] = avant.c[0];
	*s = apres.s;
}
	

FILE *OpenWriteFile(char *filename)
{
	writeFile = fopen(filename, "wb");

	return writeFile;
}	

void WriteBit (unsigned bit)
{

    if( bit )
        writeChar |= BitsMask[ writeBits % 8 ];

    if( ++writeBits % 8 )
        return;

    // fwrite(&writeChar, 1, 1, writeFile);
    putc (writeChar, writeFile);
    
    writeChar = 0;
}

void WriteBits(unsigned val, int bits)
{
	unsigned mask;

	for( mask = (1 << bits) ; bits-- ; mask >>= 1 )
		WriteBit(val & mask);
}

void CloseWriteFile()
{
	if ( writeBits % 8 )
		WriteBits( writeChar, writeBits % 8); 

	fclose(writeFile);
	writeFile = NULL;
	writeBits = 0;
}  

FILE *OpenReadFile(char *filename)
{
	readFile = fopen(filename, "ro");
	readChar = 0;

	return readFile;
}

int endReading = 0;

unsigned ReadBit ()
{
    int nread;

    if( (readBits % 16) == 0 )
    {
	// readChar = getc (readFile); // marchait pour lecture 8 bits
	nread = fread(&readChar, 1, 2, readFile);
	switch ( nread )
	{
		case 1  : fprintf(stderr, "only one byte read :( \n"); endReading=8;
		case 2  : SwapShort(&readChar); break; 
		default : fprintf(stderr, "ReadBit() pb :( \n"); endReading = 1; break;
	}
    }

    return (readChar & BitsMask[ readBits++ % 16 ] );
}

unsigned ReadBits(int bits)
{
	unsigned mask, val = 0;

	for( mask = (1 << bits) ; bits-- ; mask >>= 1 )
		val |= ReadBit();

	return val;		
}  

void CloseReadFile()
{
	fclose(readFile);
	readFile = NULL;
	readBits = 0;
}  


int main ( int argc, char **argv)
{
	FILE *file;

	unsigned short c = 0, idx;
	unsigned char *cadr = (unsigned char *)&c;

	// unsigned char nextc, prevc;
	int i, code, order, j, decal, index;
	int countBits, decBits, n, mask, nb;

	fprintf(stderr,"dehuffman v0.0v by YLP \n");

	nHuffmanCodes = 0;
	file = OpenReadFile(argv[1] ); // fopen( argv[1], "ro" );

	n = fread(HuffmanCodes, 1, sizeof(huffman_t) * 256, file);
	if( n != (256 * sizeof(huffman_t)) )
	{
		fprintf(stderr, "Pb for to read the file %s (%d bytes read) \n", argv[1], n);
		exit(0);
	}
	else
		fprintf(stderr, "header : %d bytes \n", 256 * sizeof(huffman_t) ); 
	
	for( i = 0 ; i < 256 ; i++ )
	{
		if( HuffmanCodes[i].nOccurrences && (HuffmanCodes[i].left == NULL) && (HuffmanCodes[i].right) == NULL )
		{
			code  = HuffmanCodes[i].code;
			order = HuffmanCodes[i].nBits;

			fprintf(stderr, "\ncharacter : %d (%c) [code=", HuffmanCodes[i].character, HuffmanCodes[i].character);
			PrintByte(code, order);
			fprintf(stderr, " codesize=%d] ", order);
		
		
			index = code << (16 - order);

			fprintf(stderr, "left choice=");PrintByte(index, 16);fprintf(stderr, " \n"); 	

			if ( (order >= 15) || (index > 65535L) || ( index < 0) )
			{
				fprintf(stderr, "\n codesize > 15 found :( (codesize=%d index=%d) :( \n", order, index);
				exit(0);
			}
			
/*			
			for( j = 0 ; j < (int)(pow(2.0f, 16 - order)) ; j++)
			{
				Remap[ index | j ].code      = code;
				Remap[ index | j ].nBits     = order;
				Remap[ index | j ].character = i;
				Remap[ index | j ].nOccurrences = HuffmanCodes[i].nOccurrences;				
				// PrintByte(code|j, 16); fprintf(stderr, " ");
			}
*/
			for( j = 0 ; j < (int)(pow(2, (16 - order))) ; j++ )
			{
				if ( (index + j) > (int)65535L )
				{
					// fprintf(stderr, "pb :( codesize=%d char=%d index=%d step=%d sum=%d or=%X \n", 
						// order, i, index, j, index + j, index | j);
					// exit(0);
					fprintf(stderr, ".");
					continue; 
				} 
				idx = index + j;
				// SwapShort(&idx);

				Remap[ idx ].code      = code;
				Remap[ idx ].nBits     = order;
				Remap[ idx ].character = i;
				Remap[ idx ].nOccurrences = HuffmanCodes[i].nOccurrences;				
			} 

			// fprintf(stderr,"\n"); 
		}
		// else
			// Remap[i].character = '?';
	}
			
	fprintf(stderr,"\n\n");
	
	// while ( !feof(readFile) )
	// while ( ( fread(cadr+1,1,1,file) + fread(cadr,1,1,file) ) > 1)
	while ( fread(&c, 2, 1, file) )
	{
		SwapShort(&c);
		fprintf(stderr, "\n\nfread="); PrintByte(c,16); fprintf(stderr, "\n");

		readBits += 16;
		readChar  = c;
		countBits = 0;
		// c = 0;
		// readChar = 0;

nextBits :

		// fprintf(stderr, " avant acces Remap "); PrintByte(c, 32) ; fprintf(stderr, " (c=%d) \n", c);

		while( Remap[c].nOccurrences == 0 )
		{
			fprintf(stderr,"Remap failed :( \n");
			if ( ++countBits <= 16 )
			{
				c <<= 1;
				if ( ReadBit() )
					c |= 0x01;
			}
			else
			{
				fprintf(stderr, "Bug : more than 16 bits read :( \n");
				exit(0);
			}
		}

		// fprintf(stderr, "caractere decompresse : %d (%c) [code=", Remap[c].character, Remap[c].character);
		//PrintByte(Remap[c].code,Remap[c].nBits); fprintf(stderr, "] \n");

		// PrintByte(c, 8);
		// fprintf(stderr, "\n\n%c (%d) [", Remap[c].character, Remap[c].character);
		// PrintByte( Remap[c].code >> (8 - Remap[c].nBits), Remap[c].nBits);
		// fprintf(stderr,"%d bits]    ", Remap[c].nBits);

		printf("%c", Remap[c].character);
		writeBits += 8;
		// exit(0);

/*
 		if ( feof(readFile) )
		{
			fprintf(stderr, "End of File (%d byte read %d byte write) \n", readBits / 8,  writeBits / 8);
			break;
		}
*/

		nb = Remap[c].nBits;
		c <<= nb;
		countBits += nb;
		c &= 0x0000FFFF;

		if( endReading )
		{
			fprintf(stderr, "End Of File in %d bits maximum \n", endReading);

			if ( (endReading -= nb) <= 0 )
			{ 
				fprintf(stderr, "End of File (%d byte read %d byte write) [decompression=%2.2f%%] \n", 
					readBits / 8,  writeBits / 8, (float)(100.0f * writeBits) / (float)(readBits) );
				break;
			}
			else
				goto nextBits;
		}
		else
		{
			// fprintf(stderr,"\nRead %d bits at end of ", nb); PrintByte(c,32); fprintf(stderr,"\n");
			for( n = 0, mask = 1 << (nb-1) ; n < nb ; n++, mask >>= 1)
			{
				countBits++;
				if( ReadBit() )
				{
					c |= mask;
					// fprintf(stderr,"ReadBit() work :) \n");
				}
				// PrintByte(c,32); fprintf(stderr,"(mask="); PrintByte(mask,32); fprintf(stderr,") ");
			}
			// fprintf(stderr," final="); PrintByte(c,32); fprintf(stderr,"\n");
			goto nextBits;
		}
	}

	fprintf(stderr,"FIN\n\n");
	
	return 0;
}
