#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "huffman.h"

void PrintByte(int byte, int nbits, int justify)
{
	static int pos = 0;

	int i, power;

	power = 1L << (nbits - 1);
	
	// fprintf(stderr, "nbits= %d power=%d byte=%d \n", nbits, power, byte);

	if ( justify > nbits )
	{
		for( i = 0; i < justify - nbits ; i++ )	
		{
			fprintf(stderr, " ");
			if( (++pos % 8) == 0)
			{
				fprintf(stderr, " ");
			} 
		}
	}

	for( i = 8 - nbits ; i < 8 ; i++, power >>= 1)
	{
		if( byte & power )
		{
			fprintf(stderr, "1");
		}
		else
		{
			fprintf(stderr, "0");
		}

		if( justify )
		{
			if( (++pos % 8) == 0 )
			{
				fprintf(stderr, " ");
			}
		}
		else
			pos = 0;
	}
}

int InitHuffmanCodes()
{
	int i;

	memset( HuffmanCodes, 0, sizeof(HuffmanCodes));

	for( i = 0 ; i < MAX_CHARACTERS ; i++ )
	{
		HuffmanCodes[i].character = i;
	}

	for( i = 0 ; i < MAX_NODES ; i++ )
	{
		HuffmanSorted[i] = HuffmanCodes + i;
	}

	nHuffmanCodes = MAX_CHARACTERS;
}

int ReadSymbols( FILE *file)
{

	int i;
	unsigned char c;

	while( fread(&c, 1, 1, file) == 1 )
	{
		HuffmanCodes[c].nOccurrences++;
		if ( HuffmanCodes[c].nOccurrences == 1 )
		{
			// nHuffmanCodes++;
			// HuffmanCodes[c].character = c;
			fprintf(stderr, "  Add code ASCII %d (%c) \n", c, c);
		}
	}

	rewind (file);
	fprintf(stderr, "\n\n");
} 

static int CompareHuffmanOccurences( const void  *a, const void *b)
{

	huffman_t *pa = (huffman_t *)(a);
	huffman_t *pb = (huffman_t *)(b);

/*	
	if ( pa->nOccurrences == 0 )	
		return 1;

	if ( pb->nOccurrences == 0 )	
		return -1;

	if ( pa->nOccurrences < pb->nOccurrences )
		return -1;

	if ( pa->nOccurrences > pb->nOccurrences )
		return 1;

	return 0;
*/

	return (  pa->nOccurrences - pb->nOccurrences);
} 

static int IndirectCompareHuffmanOccurences( const void  *a, const void *b)
{

	huffman_t *pa = *(huffman_t **)a;
	huffman_t *pb = *(huffman_t **)b;

	// fprintf(stderr, "indirect %p %p \n", pa, pb);

	if ( pa == NULL && pb == NULL ) return 0;

	if ( pa == NULL ) 
		return 1;

	if ( pb == NULL ) 
		return -1;

	return CompareHuffmanOccurences(pa,pb);
}

static int IndirectCompareHuffmanOccurencesInv( const void  *a, const void *b)
{
	return IndirectCompareHuffmanOccurences(b,a);
}

void ComputeHuffmanCodes( huffman_t *node, int code, int level )
{
	int i;

	// if (level > 10 ) return;

	node->code = code;
	node->nBits = level;


	if( node->left != NULL )
	{
		// fprintf(stderr, "Node left \n");
		ComputeHuffmanCodes(node->left, code * 2, level + 1);
	}
	// else fprintf(stderr, "No left \n");

	
	// if ( level && (node->code != 0) )
	{
		for( i = 0 ; i < level ; i++)
		{ 
			fprintf(stderr, "\t");
		}
		
		if( (node->left == NULL) && (node->right == NULL) )
		//if( ! (node->left && node->right) )
		{
			fprintf(stderr, "ASCII %d ", node->character);
		}
		else
		{
			fprintf(stderr, "* ");
		}
			
		fprintf(stderr, "(0x%x %dbits x%d) [", node->code, node->nBits, node->nOccurrences);
		PrintByte(node->code, node->nBits, 0); fprintf(stderr, "]\n\n");
	}

	if( node->right != NULL )
	{
		// fprintf(stderr, "Node right \n");
		ComputeHuffmanCodes(node->right, (code * 2) + 1, level + 1);
	}
	// else fprintf(stderr, "No right \n");
}

int GenerateHuffmanForest()
{

	int i, j, n;

	HuffmanRoot = HuffmanCodes + MAX_CHARACTERS;
	qsort(HuffmanSorted, nHuffmanCodes, sizeof(huffman_t *), IndirectCompareHuffmanOccurences);
	for( i = 0 ; HuffmanSorted[i]->nOccurrences == 0 ; i++);

	for( /* i = 0 , */ n = nHuffmanCodes - i ; n > 0 ; n--, i += 2 )
	{
		// fprintf(stderr, "Generate local forest of %d nodes  \n", n); 
		qsort(HuffmanSorted + i, n, sizeof(huffman_t *), IndirectCompareHuffmanOccurences);
		// fprintf(stderr, "The two less nodes [%d] and [%d] have %d and %d occurrences \n", 
			// i, i +1, HuffmanSorted[i]->nOccurrences, HuffmanSorted[i+1]->nOccurrences );

// BUG :  two bits are generated instead only one for the more frequent symbol
// TODO : handle this ...
/*
		if ( (HuffmanSorted[i]->nOccurrences == 0) || (HuffmanSorted[i+1]->nOccurrences == 0) )
		{
			fprintf(stderr, "Bug [%d]=%d [%d]=%d \n", 
				i, HuffmanSorted[i]->nOccurrences, 
				i+1, HuffmanSorted[i+1]->nOccurrences);
		}
*/				
		{ 
			HuffmanRoot=&HuffmanCodes[++nHuffmanCodes];
			HuffmanRoot->character = 255;
			HuffmanRoot->nOccurrences = HuffmanSorted[i]->nOccurrences + HuffmanSorted[i+1]->nOccurrences;

			if( HuffmanSorted[i]->nOccurrences )
				HuffmanRoot->left = HuffmanSorted[i];
			else
				HuffmanRoot->left = NULL;
				

			if( HuffmanSorted[i+1]->nOccurrences )
				HuffmanRoot->right = HuffmanSorted[i+1];
			else
				HuffmanRoot->right = NULL;

		}
		
	}

	// fprintf(stderr, "Generate Huffman codes : \n\n");
	ComputeHuffmanCodes(HuffmanRoot, 0, 0);	

}

static int CompareHuffmanCodes( const void  *a, const void *b)
{

	huffman_t *pa = (huffman_t *)(a);
	huffman_t *pb = (huffman_t *)(b);

	if( pa->nOccurrences == 0 )
		return -1;

	if( pb->nOccurrences == 0 )
		return 1;

	if ( pa->code < pb->code )
		return -1;

	if ( pa->code > pb->code )
		return 1;

	return 0;
} 

 static int IndirectCompareHuffmanCodes( const void  *a, const void *b)
{

	huffman_t *pa = *(huffman_t **)a;
	huffman_t *pb = *(huffman_t **)b;

	// fprintf(stderr, "indirect %p %p \n", pa, pb);

	if ( pa == NULL && pb == NULL ) return 0;

	if ( pa == NULL ) 
		return -1;

	if ( pb == NULL ) 
		return 1;


	return CompareHuffmanCodes(pa,pb);
}


static int CompareHuffmanCharacters( const void *a, const void *b)
{
	huffman_t *pa = (huffman_t *)(a);
	huffman_t *pb = (huffman_t *)(b);

	if ( pa->character < pb->character )
		return -1;

	if ( pa->character > pb->character )
		return 1;

	return 0;
}  

int _bits_ = 0;
unsigned char _idx_ = 0;


void WriteBits(unsigned int val, int bits)
{
	unsigned char tmp; 

	unsigned int i;

	for( i = bits ; i-- ; )
	{
		_idx_ <<= 1;
		_idx_ |= (val >> i) & 0x1;

		if( ++_bits_ == 8 )
		{
			fwrite(&_idx_, 1, 1, stdout);
			_bits_ = 0;
			_idx_ = 0;
		}
	}
}

void FinishWriteBits()
{
	unsigned char tmp;

	if( _bits_ % 8)
	{
		_idx_ <<= (_bits_ % 8);
		fwrite(&tmp, 1, 1, stdout);
		_bits_= 0;
		_idx_ = 0;
	} 
}	

int main( int argc, char *argv[] )
{
	FILE *file;
	int i, n, bits;
	unsigned char c;
	


	fprintf(stderr, "\nHuffman compression by YLP v%d.%d \n\n", MAJOR_VERSION, MINOR_VERSION);

	if( (file = fopen(argv[1], "ro")) == NULL )
	{
		fprintf(stderr, "cannot open %s :( \n", argv[1] );
		return 0;
	}
		 

	// fprintf(stderr, "Init HuffmanCodes \n\n");
	InitHuffmanCodes();

	fprintf(stderr, "Read symbols : \n\n");
	ReadSymbols(file);

/*
	qsort(HuffmanCodes, 256, sizeof(huffman_t), CompareHuffmanOccurences);
	for( i = 0 ; HuffmanCodes[i].nOccurrences ; i++, HuffmanRoot++ )
	{
		fprintf(stderr, "HuffmanCodes[%d] : character=%c nOccurrences=%d \n", 
			i, HuffmanCodes[i].character, HuffmanCodes[i].nOccurrences );
	}
	
	fprintf(stderr, "\n\n");
*/

	
	fprintf(stderr, "Generate Huffman forest : \n\n");
	GenerateHuffmanForest();

	fprintf(stderr, "\n\nGenerated codes : \n\n");
	qsort(HuffmanSorted, nHuffmanCodes, sizeof(huffman_t *), IndirectCompareHuffmanOccurencesInv );
	for( i = 0 , n = 0, bits = 0 ; i < nHuffmanCodes  ; i++ )
	{
		if( (HuffmanSorted[i]->left == NULL) && (HuffmanSorted[i]->right == NULL) && HuffmanSorted[i]->nOccurrences)
		{
			PrintByte(HuffmanSorted[i]->code, HuffmanSorted[i]->nBits, 8);
			fprintf(stderr, " (ASCII=%d [%c] code=%d nOccurences=%-2d  bits/symbol=%-2d  total=%d bits) \n",
				HuffmanSorted[i]->character,
				HuffmanSorted[i]->character,
				HuffmanSorted[i]->code,
				HuffmanSorted[i]->nOccurrences, 
				HuffmanSorted[i]->nBits,
				HuffmanSorted[i]->nOccurrences * HuffmanSorted[i]->nBits				
			);


			n    += HuffmanSorted[i]->nOccurrences;
			bits += HuffmanSorted[i]->nBits * HuffmanSorted[i]->nOccurrences;
		}
	}

/*
	if( bits % 8 )
	{
		bits += 8 - (bits % 8);
	}
*/

	

/*

	// qsort(HuffmanCodes, nHuffmanCodes, sizeof(huffman_t), CompareHuffmanCharacters);

	for( i = 0 ; i < MAX_CHARACTERS ; i++)
	{
		if( HuffmanCodes[i].nOccurrences )
		{
			fprintf(stderr, "HuffmanCodes[%3d] : ASCII=%3d left=0x%p right=0x%p nOccurences=%d code=%d nbits=%d \n",
				i,
				HuffmanCodes[i].character,
				HuffmanCodes[i].left,
				HuffmanCodes[i].right,
				HuffmanCodes[i].nOccurrences,
				HuffmanCodes[i].code,
				HuffmanCodes[i].nBits		
			);
		}	
	}
*/


	fprintf(stderr, "\n\nCoded string : \n\n");
	bits = 0;
	fwrite(HuffmanCodes, sizeof(huffman_t), 256, stdout);
	while( fread(&c, 1, 1, file) == 1 )
	{
		fprintf(stderr, "(%c) ", c); 
		PrintByte( HuffmanCodes[c].code, HuffmanCodes[c].nBits, 0);
		bits += HuffmanCodes[c].nBits;
		WriteBits(HuffmanCodes[c].code, HuffmanCodes[c].nBits);
		fprintf(stderr, "\n");
	}
	fclose(file);
	FinishWriteBits();

	// bits -> byte end padding
	fprintf(stderr, "[");
	while( bits % 8 )
	{
		fprintf(stderr, "0");
		bits++;
	}
	fprintf(stderr, "]");

	fprintf(stderr, "\n\nCompression : %d/%d = %2.2f%% \n\n", bits/8, n , (float)(12.5f * bits) / (float)(n) );			

	return 0;
}
